# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
CUDA
----

Applications that need to perform massively parallel operations, like
processing large arrays, may use the CUDA framework to accelerate their
processing on graphics processing units (GPUs). The CUDA framework has two
complementary pieces to it.

There are a set of GPU instruction set extensions that are implemented by
various graphics cards. These instruction set extensions are known as the CUDA
Compute Capability.

The second part of the framework is an SDK that allows developers to take
advantage of the hardware's instruction set extensions of a particular version
(a specific CUDA Compute Capability version, that is).

An application will link with a version of the CUDA SDK, and the version of the
CUDA SDK controls which CUDA Compute Capability versions the application will
be able to work with.

The ``os_traits.hw.gpu.cuda`` module contains traits for both the CUDA compute
capability version as well as the CUDA SDK version. For example,
``os_traits.hw.gpu.cuda.COMPUTE_CAPABILITY_V3_2`` and
``os_traits.hw.gpu.cuda.SDK_V6_5``.

The ``os_traits.hw.gpu.cuda`` module contains a utility function called
``compute_capabilities_supported()`` that accepts a trait indicating the CUDA
SDK version and returns a ``set()`` containing the matching CUDA compute
capability traits that that version of the CUDA SDK knows how to utilize.

Here is an example of listing the CUDA compute capability version traits that
the CUDA SDK 8.0 is capable of working with::

    >>> from os_traits.hw.gpu import cuda
    >>> import pprint
    >>>
    >>> sdk8_caps = cuda.compute_capabilities_supported(cuda.SDK_V8_0)
    >>> pprint.pprint(sdk8_caps)
    set(['HW_GPU_CUDA_COMPUTE_CAPABILITY_V2_0',
         'HW_GPU_CUDA_COMPUTE_CAPABILITY_V2_1',
         'HW_GPU_CUDA_COMPUTE_CAPABILITY_V3_0',
         'HW_GPU_CUDA_COMPUTE_CAPABILITY_V3_2',
         'HW_GPU_CUDA_COMPUTE_CAPABILITY_V3_5',
         'HW_GPU_CUDA_COMPUTE_CAPABILITY_V3_7',
         'HW_GPU_CUDA_COMPUTE_CAPABILITY_V5_0',
         'HW_GPU_CUDA_COMPUTE_CAPABILITY_V5_2',
         'HW_GPU_CUDA_COMPUTE_CAPABILITY_V5_3',
         'HW_GPU_CUDA_COMPUTE_CAPABILITY_V6_0',
         'HW_GPU_CUDA_COMPUTE_CAPABILITY_V6_1',
         'HW_GPU_CUDA_COMPUTE_CAPABILITY_V6_2'])

For more information on CUDA, see the `Wikipedia article`_.

.. _Wikipedia article: https://en.wikipedia.org/wiki/CUDA
"""


TRAITS = [
    # ref: https://en.wikipedia.org/wiki/CUDA
    # ref: https://developer.nvidia.com/cuda-toolkit-archive
    'COMPUTE_CAPABILITY_V1_0',
    'COMPUTE_CAPABILITY_V1_1',
    'COMPUTE_CAPABILITY_V1_2',
    'COMPUTE_CAPABILITY_V1_3',
    'COMPUTE_CAPABILITY_V2_0',
    'COMPUTE_CAPABILITY_V2_1',
    'COMPUTE_CAPABILITY_V3_0',
    'COMPUTE_CAPABILITY_V3_2',
    'COMPUTE_CAPABILITY_V3_5',
    'COMPUTE_CAPABILITY_V3_7',
    'COMPUTE_CAPABILITY_V5_0',
    'COMPUTE_CAPABILITY_V5_2',
    'COMPUTE_CAPABILITY_V5_3',
    'COMPUTE_CAPABILITY_V6_0',
    'COMPUTE_CAPABILITY_V6_1',
    'COMPUTE_CAPABILITY_V6_2',
    'COMPUTE_CAPABILITY_V7_0',
    'COMPUTE_CAPABILITY_V7_1',
    'COMPUTE_CAPABILITY_V7_2',
    'SDK_V6_5',
    'SDK_V7_5',
    'SDK_V8_0',
    'SDK_V9_0',
    'SDK_V9_1',
    'SDK_V9_2',
    'SDK_V10_0',
]

_CAPS_V1 = [
    'HW_GPU_CUDA_COMPUTE_CAPABILITY_V1_0',
    'HW_GPU_CUDA_COMPUTE_CAPABILITY_V1_1',
    'HW_GPU_CUDA_COMPUTE_CAPABILITY_V1_2',
    'HW_GPU_CUDA_COMPUTE_CAPABILITY_V1_3',
]

_CAPS_V2 = [
    'HW_GPU_CUDA_COMPUTE_CAPABILITY_V2_0',
    'HW_GPU_CUDA_COMPUTE_CAPABILITY_V2_1',
]

_CAPS_V3 = [
    'HW_GPU_CUDA_COMPUTE_CAPABILITY_V3_0',
    'HW_GPU_CUDA_COMPUTE_CAPABILITY_V3_2',
    'HW_GPU_CUDA_COMPUTE_CAPABILITY_V3_5',
    'HW_GPU_CUDA_COMPUTE_CAPABILITY_V3_7',
]

_CAPS_V5 = [
    'HW_GPU_CUDA_COMPUTE_CAPABILITY_V5_0',
    'HW_GPU_CUDA_COMPUTE_CAPABILITY_V5_2',
    'HW_GPU_CUDA_COMPUTE_CAPABILITY_V5_3',
]

_CAPS_V6 = [
    'HW_GPU_CUDA_COMPUTE_CAPABILITY_V6_0',
    'HW_GPU_CUDA_COMPUTE_CAPABILITY_V6_1',
    'HW_GPU_CUDA_COMPUTE_CAPABILITY_V6_2',
]

_CAPS_V7 = [
    'HW_GPU_CUDA_COMPUTE_CAPABILITY_V7_0',
    'HW_GPU_CUDA_COMPUTE_CAPABILITY_V7_1',
    'HW_GPU_CUDA_COMPUTE_CAPABILITY_V7_2',
]

_SDK_COMPUTE_CAP_SUPPORT = {
    'HW_GPU_CUDA_SDK_V6_5': set(
        _CAPS_V1 + _CAPS_V2 + _CAPS_V3 + _CAPS_V5
    ),
    'HW_GPU_CUDA_SDK_V7_5': set(
        _CAPS_V2 + _CAPS_V3 + _CAPS_V5
    ),
    'HW_GPU_CUDA_SDK_V8_0': set(
        _CAPS_V2 + _CAPS_V3 + _CAPS_V5 + _CAPS_V6
    ),
    'HW_GPU_CUDA_SDK_V9_0': set(
        _CAPS_V3 + _CAPS_V5 + _CAPS_V6 + _CAPS_V7
    ),
    'HW_GPU_CUDA_SDK_V9_1': set(
        _CAPS_V3 + _CAPS_V5 + _CAPS_V6 + _CAPS_V7
    ),
    'HW_GPU_CUDA_SDK_V9_2': set(
        _CAPS_V3 + _CAPS_V5 + _CAPS_V6 + _CAPS_V7
    ),
    'HW_GPU_CUDA_SDK_V10_0': set(
        _CAPS_V3 + _CAPS_V5 + _CAPS_V6 + _CAPS_V7
    ),
}


def compute_capabilities_supported(sdk_trait):
    """Given an SDK trait, returns a set of compute capability traits that the
    version of the SDK supports.

    Returns None if no matches were found for the SDK trait.
    """
    return _SDK_COMPUTE_CAP_SUPPORT.get(sdk_trait)



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


TRAITS = [
    # ref: https://en.wikipedia.org/wiki/Streaming_SIMD_Extensions
    'AVX',
    'AVX2',
    'CLMUL',
    'FMA3',
    'FMA4',
    'F16C',
    'MMX',
    'SSE',
    'SSE2',
    'SSE3',
    'SSSE3',
    'SSE41',
    'SSE42',
    'SSE4A',
    'XOP',
    '3DNOW',

    # ref: https://en.wikipedia.org/wiki/AVX-512
    'AVX512F',  # foundation
    'AVX512CD',  # conflict detection
    'AVX512PF',  # prefetch
    'AVX512ER',  # exponential + reciprocal
    'AVX512VL',  # vector length extensions
    'AVX512BW',  # byte + word
    'AVX512DQ',  # double word + quad word
    'AVX512VNNI',  # vector neural network instructions
    'AVX512VBMI',  # vector byte manipulation instructions
    'AVX512IFMA',  # integer fused multiply add
    'AVX512VBMI2',  # vector byte manipulation instructions 2
    'AVX512BITALG',  # bit algorithms
    'AVX512VAES',  # vector aes instructions
    'AVX512GFNI',  # galois field new instructions
    'AVX512VPCLMULQDQ',  # carry-less multiplication of quadwords
    'AVX512VPOPCNTDQ',  # vector population count instruction
    # ref: https://en.wikipedia.org/wiki/Bit_Manipulation_Instruction_Sets
    'ABM',
    'BMI',
    'BMI2',
    'TBM',
    # ref: https://en.wikipedia.org/wiki/AES_instruction_set
    'AESNI',
    # ref: https://en.wikipedia.org/wiki/Intel_SHA_extensions
    'SHA',
    # ref: https://en.wikipedia.org/wiki/Intel_MPX
    'MPX',
    # ref: https://en.wikipedia.org/wiki/Software_Guard_Extensions
    'SGX',
    # ref:
    #    https://en.wikipedia.org/wiki/Transactional_Synchronization_Extensions
    'TSX',
    # ref: https://en.wikipedia.org/wiki/Advanced_Synchronization_Facility
    'ASF',
    # ref: https://en.wikipedia.org/wiki/VT-x
    # NOTE(kchamart): The 'VMX' trait is Intel-only, and does not belong
    # in this file (which is supposed to be a "common" file for all
    # x86-related).  But we need to retain it here forever to not cause
    # Placement breakage.
    'VMX',
    # ref: https://en.wikipedia.org/wiki/AMD-V
    # NOTE(kchamart): The 'SVM' trait is AMD-only, and does not belong
    # in this "common" file.  But we need to retain it here forever to
    # not cause Placement breakage.
    'SVM',
    # ref: https://git.qemu.org/?p=qemu.git;a=blob;f=docs/qemu-cpu-models.texi
    # Recommended to allow guest OS to use 1 GB size memory pages.  Not
    # included by default in any of the Intel and AMD CPU models.  So
    # this should be explicitly turned on for all Intel and AMD CPU
    # models.
    'PDPE1GB',
    # ref: https://git.qemu.org/?p=qemu.git;a=blob;f=docs/qemu-cpu-models.texi
    # Required to enable stronger Spectre v2 (CVE-2017-5715) fixes in
    # some operating systems.  This flag must be explicitly turned on
    # for *all* Intel and AMD CPU models.  (Prerequisite: host CPU
    # microcode needs to support this feature before it can be used for
    # guest CPUs).
    'STIBP',
]



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Characteristics of PCI devices
TRAITS = [
    # PCI device can be live-migrated from one compute node to another
    # with the same device.
    'LIVE_MIGRATABLE',
    # PCI device lifecycle is being managed as "one time use". Compute manager
    # will set reserved=total during assignment and leave it as such during
    # deallocation, requiring an external agent to un-reserve it before it
    # can be used again.
    'ONE_TIME_USE',
]



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


TRAITS = [
    # ref: https://docs.openstack.org/os-traits/latest/contributor/index.html#trait-lifecycle-policy # noqa
    # NOTE(kchamart): This file is deprecated.  The 'SEV' trait is
    # AMD-only, so it is copied to hw/cpu/amd.py; it is retained here
    # not to cause Placement breakage.  All AMD-only traits are being
    # tracked under: hw/cpu/x86/amd.py.  And the traits common to both
    # AMD _and_ Intel are being tracked here: hw/cpu/x86/__init__.py.
    #
    # NOTE(aspiers): This trait was never used for anything, since the
    # first bit of SEV code to use an SEV trait will land after this
    # https://review.opendev.org/#/c/638680/ which has an explicit
    # 'Depends-On' against the change I1c9a72d19ef ("hw: cpu: Rework the
    # directory layout; add missing traits"), and is actually blocked
    # until I1c9a72d19ef merges *and* gets released.
    'SEV',
]



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

TRAITS = [
    # The virt driver supports attaching a volume after boot
    'ATTACH',
    # The virt driver supports attaching a volume after boot and specifying a
    # device tag for the volume
    'ATTACH_WITH_TAG',
    # The virt driver supports extending a volume after boot
    'EXTEND',
    # The virt driver supports volumes that can be attached to multiple guests
    'MULTI_ATTACH',
]



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import importlib
import pkgutil
import re
import sys

import pbr.version

THIS_NAME = __name__
THIS_LIB = sys.modules[THIS_NAME]
TEST_DIR = "%s.tests" % THIS_NAME

__version__ = pbr.version.VersionInfo(THIS_NAME).version_string()

# Any user-specified feature/trait is prefixed with the custom namespace
CUSTOM_NAMESPACE = 'CUSTOM_'


def _symbolize(mod_name, props):
    """Given a reference to a Python module object and an iterable of short
    string names for traits, registers symbols in the module corresponding to
    the full namespaced name for each trait.
    """
    for prop in props:
        leaf_mod = sys.modules[mod_name]
        value_base = '_'.join([m.upper() for m in mod_name.split('.')[1:]])
        value = value_base + '_' + prop.upper()
        setattr(THIS_LIB, value, value)  # os_traits.HW_CPU_X86_SSE
        setattr(leaf_mod, prop, value)  # os_traits.hw.cpu.x86.SSE


def _visualize(mod_name, props, seen=None):
    if mod_name in seen:
        return
    seen.add(mod_name)
    components = mod_name.split('.')
    tab = '   '
    # Print the module name
    indent = tab * (len(components) - 1)
    print('%s%s:' % (indent, components[-1].upper()))
    # Print the properties
    indent = tab * len(components)
    if props:
        print('%s%s' % (indent, ', '.join(props)))


def _walk_submodules(package, recursive, callback, **kwargs):
    """Recursively walk the repository's submodules and invoke a callback for
    each module with the list of short trait names found therein.

    :param package: The package (name or module obj) to start from.
    :param recursive: If True, recurse depth-first.
    :param callback: Callable to be invoked for each module. The signature is::

            callback(mod_name, props, **kwargs)

        * mod_name: the string name of the module (e.g. 'os_traits.hw.cpu').
        * props: an iterable of short string names for traits, gleaned from the
          TRAITS member of that module, defaulting to [].
        * kwargs: The same kwargs as passed to _walk_submodules, useful for
          tracking data across calls.
    :param kwargs: Arbitrary keyword arguments to be passed to the callback on
        each invocation.
    """
    if isinstance(package, str):
        package = importlib.import_module(package)
    for loader, mod_name, is_pkg in pkgutil.walk_packages(
            package.__path__, package.__name__ + '.'):
        if TEST_DIR in mod_name:
            continue
        imported = importlib.import_module(mod_name)
        props = getattr(imported, "TRAITS", [])
        callback(mod_name, props, **kwargs)
        if recursive and is_pkg:
            _walk_submodules(mod_name, recursive, callback, **kwargs)


# This is where the names defined in submodules are imported by recursively
# importing all submodules/subpackages and symbolizing their TRAITS
_walk_submodules(sys.modules.get(__name__), True, _symbolize)


def get_traits(prefix=None, suffix=None):
    """Returns the trait strings in the os_traits module, optionally filtered
    by a supplied prefix and suffix.

    :param prefix: Optional string prefix to filter by. e.g. 'HW_'
    :param suffix: Optional string suffix to filter by, e.g. 'SSE'
    """
    prefix = prefix or ""
    suffix = suffix or ""
    return [
        v for k, v in sys.modules[__name__].__dict__.items()
        if isinstance(v, str) and
        not k.startswith('_') and
        v.startswith(prefix) and
        v.endswith(suffix) and
        # skip module constants
        k not in ('CUSTOM_NAMESPACE', 'THIS_NAME', 'THIS_LIB', 'TEST_DIR')
    ]


def check_traits(traits, prefix=None):
    """Returns a tuple of two trait string sets, the first set contains valid
    traits, and the second contains others.

    :param traits: An iterable contains trait strings.
    :param prefix: Optional string prefix to filter by. e.g. 'HW_'
    """
    trait_set = set(traits)
    valid_trait_set = set(get_traits(prefix))

    valid_traits = trait_set & valid_trait_set

    return (valid_traits, trait_set - valid_traits)


def is_custom(trait):
    """Returns True if the trait string represents a custom trait, or False
    otherwise.

    :param trait: String name of the trait
    """
    return trait.startswith(CUSTOM_NAMESPACE)


def normalize_name(name):
    """Converts an input string to a legal* custom trait name.

    Legal custom trait names are prefixed with CUSTOM_ and contain only the
    characters A-Z, 0-9, and _ (underscore).

    *Does not attempt to handle length restrictions.

    :param name: A string to be converted.
    :return: A legal* custom trait name.
    """
    if name is None:
        return None
    # Replace non-alphanumeric characters with underscores
    norm_name = re.sub('[^0-9A-Za-z]+', '_', name)
    # Bug #1762789: Do .upper after replacing non alphanumerics.
    return CUSTOM_NAMESPACE + norm_name.upper()


def print_tree():
    """Print (to stdout) a visual representation of all the namespaces and the
    (short) trait names defined therein.

    """
    _walk_submodules(sys.modules.get(__name__), True, _visualize, seen=set())



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

TRAITS = [
    # Traits corresponding to the allowed values of "hw_usb_model"
    # image metadata property. These are used by the nova libvirt driver
    # however we have decided for now to not support the other older options
    # for performance and support reasons. Options are documented at
    # https://libvirt.org/formatdomain.html#controllers
    'MODEL_QEMU_XHCI',
    'MODEL_NEC_XHCI'
]
